//##################################################################################################
//
//  Ceetron Data Provider Framework
//
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2018, Ceetron AS
//##################################################################################################

#include "cdphlChangeNotifications.h"
#include "cdphlAssert.h"
#include <algorithm>

namespace cdphl {
    
//==================================================================================================
//
//
//
//==================================================================================================


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
struct StateInfoLesser
{
    inline bool operator() (const StateInfo* state1, const StateInfo* state2)
    {
        return (state1->referenceValue() < state2->referenceValue());
    }
};


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
ChangeNotifications::ChangeNotifications()
: rigidBodyTransformationsChanged(false)
{
}
    

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
ChangeNotifications::~ChangeNotifications()
{
    for (size_t i = 0; i < resultAddedInfoArr.size(); ++i)
    {
        delete resultAddedInfoArr[i];
    }
    
    for (size_t i = 0; i < stateMetaDataChangedInfoArr.size(); ++i)
    {
        delete stateMetaDataChangedInfoArr[i];
    }
    
    for (size_t i = 0; i < stateAddedInfoArr.size(); ++i)
    {
        delete stateAddedInfoArr[i];
    }
}

    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ChangeNotifications::notifyResultChanged(cdp_int_t resultId)
{
    // Verify valid and not added or deleted result
    if (resultIdExists(resultId))
    {
        return;
    }
    
    // Register it
    resultChangedIdArr.push_back(resultId);
}
    
    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
CDPResultInfo* ChangeNotifications::notifyResultAdded(CDPResultType resultType, cdp_int_t resultId, const CDPString& name, CDPResultMappingType mapping)
{
    // Verify valid and not added or deleted result
    if (resultIdExists(resultId))
    {
        return NULL;
    }
    
    // Register it
    ResultInfo* resultInfo = new ResultInfo(resultType, resultId, name, mapping);
    resultAddedInfoArr.push_back(resultInfo);
    
    return resultInfo;
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ChangeNotifications::notifyResultDeleted(cdp_int_t resultId)
{
    // Verify valid and not added or deleted result
    if (resultIdExists(resultId))
    {
        return;
    }
    
    // Register it
    resultDeletedIdArr.push_back(resultId);
}

    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ChangeNotifications::notifyRigidBodyTransformationsChanged()
{
    rigidBodyTransformationsChanged = true;
}

    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ChangeNotifications::notifyStateMetaDataChanged(cdp_int_t stateId, const CDPString& name, cdp_double_t referenceValue)
{
    // Verify valid and not added or deleted result
    if (stateIdExists(stateId))
    {
        return;
    }
    
    // Register it
    StateInfo* stateInfo = new StateInfo(stateId, name, referenceValue);
    stateMetaDataChangedInfoArr.push_back(stateInfo);
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
CDPStateInfo* ChangeNotifications::notifyStateAdded(cdp_int_t stateId, const CDPString& name, cdp_double_t referenceValue)
{
    // Verify valid and not added or deleted result
    if (stateIdExists(stateId))
    {
        return NULL;
    }
    
    // Register it
    StateInfo* stateInfo = new StateInfo(stateId, name, referenceValue);
    stateAddedInfoArr.push_back(stateInfo);

    std::sort(stateAddedInfoArr.begin(),stateAddedInfoArr.end(), StateInfoLesser());
   
    return stateInfo;
}
    

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ChangeNotifications::notifyStateDeleted(cdp_int_t stateId)
{
    // Verify valid and not added or deleted result
    if (stateIdExists(stateId))
    {
        return;
    }
    
    // Register it
    stateDeletedIdArr.push_back(stateId);

    std::sort(stateDeletedIdArr.begin(), stateDeletedIdArr.end());
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ChangeNotifications::notifyGeometryChanged(cdp_size_t geometryIndex)
{
    // Verify valid and not added or deleted result
    for (size_t i = 0; i < geometryChangedIndexArr.size(); ++i)
    {
        if (geometryChangedIndexArr[i] == geometryIndex)
        {
            return;
        }
    }
    
    // Register it
    geometryChangedIndexArr.push_back(geometryIndex);
}

    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
bool ChangeNotifications::resultIdExists(cdp_int_t resultId) const
{
    for (size_t i = 0; i < resultChangedIdArr.size(); ++i)
    {
        if (resultChangedIdArr[i] == resultId)
        {
            return true;
        }
    }
    
    for (size_t i = 0; i < resultAddedInfoArr.size(); ++i)
    {
        if (resultAddedInfoArr[i]->id() == resultId)
        {
            return true;
        }
    }
    
    for (size_t i = 0; i < resultDeletedIdArr.size(); ++i)
    {
        if (resultDeletedIdArr[i] == resultId)
        {
            return true;
        }
    }

    return false;
}
    

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
bool ChangeNotifications::stateIdExists(cdp_int_t stateId) const
{
    for (size_t i = 0; i < stateAddedInfoArr.size(); ++i)
    {
        if (stateAddedInfoArr[i]->id() == stateId)
        {
            return true;
        }
    }
    
    for (size_t i = 0; i < stateMetaDataChangedInfoArr.size(); ++i)
    {
        if (stateMetaDataChangedInfoArr[i]->id() == stateId)
        {
            return true;
        }
    }
    
    for (size_t i = 0; i < stateDeletedIdArr.size(); ++i)
    {
        if (stateDeletedIdArr[i] == stateId)
        {
            return true;
        }
    }
    
    return false;
}


}  // cdphl
