//##################################################################################################
//
//  Ceetron Data Provider Framework
//
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2018, Ceetron AS
//##################################################################################################

#include "cdphlElementGroup.h"
#include "cdphlAssert.h"

#include <set>

namespace cdphl {


//==================================================================================================
//
//
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
size_t ElementSubGroupData::computeDerivedData(const std::vector<cdp_ubyte_t>& polyhedronNumNodesPerElement, size_t polyhedronElementIndexOffset)
{
    elementStartArr.resize(elementCount);
    
    int eltStartIdx = 0;

    for (size_t i = 0; i < elementCount; ++i)
    {
        elementStartArr[i] = eltStartIdx;
        
        CDPElementType eltType = elementTypeArr.size() > 0 ? elementTypeArr[i] : elementType;
        
        if (eltType == CDP_ELT_POLYHEDRONS)
        {
            CDPHL_ASSERT(polyhedronElementIndexOffset < polyhedronNumNodesPerElement.size());
            eltStartIdx += polyhedronNumNodesPerElement[polyhedronElementIndexOffset];
            polyhedronElementIndexOffset++;
        }
        else
        {
            eltStartIdx += static_cast<int>(ElementGroup::fixedNodeCountForElement(eltType));
        }
    }
    
    return polyhedronElementIndexOffset;
}

    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
int ElementSubGroupData::numElementNodes(size_t elementIndex) const
{
    // If not, computeDerivedData is probably not done
    CDPHL_ASSERT(elementStartArr.size() == elementCount);
    
    if (elementIndex == elementCount - 1)
    {
        return static_cast<int>(elementNodeArr.size()) - elementStartArr[elementIndex];
    }
    
    return elementStartArr[elementIndex + 1] - elementStartArr[elementIndex];
}

//==================================================================================================
//
//
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
ElementGroup::ElementGroup(cdp_size_t nodeGroupIndex)
{
    m_nodeGroupIndex = nodeGroupIndex;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
ElementGroup::~ElementGroup()
{
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ElementGroup::addSingleTypeElements(CDPElementType elementType, cdp_size_t elementCount, const cdp_int_t* elementNodeArr, cdp_size_t elementNodeCount, const cdp_int_t* elementIdArr)
{
    m_subGroupDataArr.push_back(ElementSubGroupData());
    ElementSubGroupData& group = m_subGroupDataArr.back();

    group.elementType = elementType;
    group.elementCount = elementCount;

    group.elementNodeArr.reserve(elementNodeCount);
    group.elementNodeArr.assign(elementNodeArr, elementNodeArr + elementNodeCount);
        
    if (elementIdArr && elementCount > 0)
    {
        group.elementIdArr.reserve(elementCount);
        group.elementIdArr.assign(elementIdArr, elementIdArr + elementCount);
    }
    
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ElementGroup::addMultipleTypeElements(const CDPElementType* elementTypeArr, cdp_size_t elementCount, const cdp_int_t* elementNodeArr, cdp_size_t elementNodeCount, const cdp_int_t* elementIdArr)
{
    
    m_subGroupDataArr.push_back(ElementSubGroupData());
    ElementSubGroupData& group = m_subGroupDataArr.back();
  
    group.elementType = CDP_ELT_UNKNOWN;
    group.elementCount = elementCount;

    group.elementTypeArr.reserve(elementCount);
    group.elementTypeArr.assign(elementTypeArr, elementTypeArr + elementCount);

    group.elementNodeArr.reserve(elementNodeCount);
    group.elementNodeArr.assign(elementNodeArr, elementNodeArr + elementNodeCount);
    
    if (elementIdArr && elementCount > 0)
    {
        group.elementIdArr.reserve(elementCount);
        group.elementIdArr.assign(elementIdArr, elementIdArr + elementCount);
    }
    
    
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ElementGroup::addPolyhedronData(const cdp_ubyte_t* numNodesPerPolyhedronArr, const cdp_ubyte_t* numFacesPerPolyhedronArr, cdp_size_t polyhedronCount, const cdp_ubyte_t* numNodesPerFaceArr, cdp_size_t totalFaceCount, const cdp_int_t* faceNodeIndicesArr, cdp_size_t faceNodeIndicesCount)
{
    m_polyhedronNumNodesPerElement.reserve(m_polyhedronNumNodesPerElement.size() + polyhedronCount);
    m_polyhedronNumNodesPerElement.insert(m_polyhedronNumNodesPerElement.end(), numNodesPerPolyhedronArr, numNodesPerPolyhedronArr + polyhedronCount);
    
    m_polyhedronFaceCountArr.reserve(polyhedronCount);
    m_polyhedronFaceCountArr.assign(numFacesPerPolyhedronArr, numFacesPerPolyhedronArr + polyhedronCount);
    
    m_polyhedronFaceNodeCountArr.reserve(totalFaceCount);
    m_polyhedronFaceNodeCountArr.assign(numNodesPerFaceArr, numNodesPerFaceArr + totalFaceCount);
    
    m_polyhedronFaceNodeIndicesArr.reserve(m_polyhedronFaceNodeIndicesArr.size() + faceNodeIndicesCount);
    m_polyhedronFaceNodeIndicesArr.insert(m_polyhedronFaceNodeIndicesArr.end(), faceNodeIndicesArr, faceNodeIndicesArr + faceNodeIndicesCount);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ElementGroup::computeDerivedData()
{
    size_t polyhedronOffset = 0;
    for (size_t i = 0; i < m_subGroupDataArr.size(); ++i)
    {
        polyhedronOffset += m_subGroupDataArr[i].computeDerivedData(m_polyhedronNumNodesPerElement, polyhedronOffset);
    }
}
    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ElementGroup::setElementPartIds(const cdp_int_t* partIdArr, cdp_size_t elementCount)
{
    m_elementPartIdArr.reserve(elementCount);
    m_elementPartIdArr.assign(partIdArr, partIdArr + elementCount);
}
    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ElementGroup::setPrescribedElementOrdering(const cdp_int_t* elementTargetIndices, cdp_size_t elementCount)
{
    m_prescribedElementIndices.reserve(elementCount);
    m_prescribedElementIndices.assign(elementTargetIndices, elementTargetIndices + elementCount);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
size_t ElementGroup::nodeGroupIndex() const
{
    return m_nodeGroupIndex;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
size_t ElementGroup::subGroupDataCount() const
{
    return m_subGroupDataArr.size();
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
const ElementSubGroupData& ElementGroup::subGroupData(size_t index) const
{
    return m_subGroupDataArr[index];
}
    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
const std::vector<cdp_ubyte_t>& ElementGroup::polyhedronFaceCountArr() const
{
    return m_polyhedronFaceCountArr;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
const std::vector<cdp_ubyte_t>& ElementGroup::polyhedronFaceNodeCountArr() const
{
    return m_polyhedronFaceNodeCountArr;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
const std::vector<cdp_int_t>& ElementGroup::polyhedronFaceNodeIndicesArr() const
{
    return m_polyhedronFaceNodeIndicesArr;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
const std::vector<cdp_int_t>& ElementGroup::elementPartIdArr() const
{
    return m_elementPartIdArr;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
const std::vector<cdp_int_t>& ElementGroup::prescribedElementIndices() const
{
    return m_prescribedElementIndices;
}

    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
size_t ElementGroup::elementCount() const
{
    size_t eltCount = 0;
    for (size_t i = 0; i < m_subGroupDataArr.size(); ++i)
    {
        eltCount += m_subGroupDataArr[i].elementCount;
    }
    
    return eltCount;
}

    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
size_t ElementGroup::fixedNodeCountForElement(CDPElementType elementType)
{
    CDPHL_ASSERT(elementType != CDP_ELT_POLYHEDRONS);

    switch (elementType) 
    {
        case CDP_ELT_UNKNOWN:           return 0;
        case CDP_ELT_BEAMS:             return 2;
        case CDP_ELT_BEAMS_3:           return 3;
        case CDP_ELT_TRIANGLES:         return 3;
        case CDP_ELT_TRIANGLES_6:       return 6;
        case CDP_ELT_QUADS:             return 4;
        case CDP_ELT_QUADS_8:           return 8;
        case CDP_ELT_QUADS_9:           return 9;
        case CDP_ELT_TETRAHEDRONS:      return 4;
        case CDP_ELT_TETRAHEDRONS_10:   return 10;
        case CDP_ELT_HEXAHEDRONS:       return 8;
        case CDP_ELT_HEXAHEDRONS_20:    return 20;
        case CDP_ELT_PENTAHEDRONS:      return 6;
        case CDP_ELT_PENTAHEDRONS_15:   return 15;
        case CDP_ELT_POINTS:            return 1;
        case CDP_ELT_PYRAMIDS:          return 5;
        case CDP_ELT_PYRAMIDS_13:       return 13;

        case CDP_ELT_POLYHEDRONS:       return CDP_UNDEFINED_SIZE_T;
    }

    return 0;
}


}  // cdphl
