//##################################################################################################
//
//  Ceetron Data Provider Framework
//
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2018, Ceetron AS
//##################################################################################################

#include "cdphlMetaData.h"
#include "cdphlAssert.h"

namespace cdphl {

//==================================================================================================
//
//
//
//==================================================================================================
SimulationInfo::SimulationInfo()
{
    simulationType = CDP_SIMULATION_TYPE_OTHER;
    solutionType = CDP_SOLUTION_TYPE_OTHER;
}

//==================================================================================================
//
//
//
//==================================================================================================
ElementGroupInfo::ElementGroupInfo(cdp_int_t defaultPartId, const CDPString& defaultPartName)
 :  m_defaultPartId(defaultPartId),
    m_defaultPartName(defaultPartName)
{
}

///--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ElementGroupInfo::setPartName(cdp_int_t geometryGlobalPartId, const CDPString& partName)
{
    m_partInfoArr.push_back(PartInfo(geometryGlobalPartId, partName));
}


///--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ElementGroupInfo::addPartData(cdp_int_t geometryGlobalPartId, const CDPString& key, const CDPString& value)
{
    for (size_t i = 0; i < m_partInfoArr.size(); ++i)
    {
        if (m_partInfoArr[i].id == geometryGlobalPartId)
        {
            std::string skey = key.utf8();
            std::string svalue = value.utf8();
            m_partInfoArr[i].partData[skey] = svalue;
            return;
        }
    }
    CDPHL_FAIL_MSG("geometryGlobalPartId not found");
}


///--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ElementGroupInfo::setPartBoundaryCondition(cdp_int_t geometryGlobalPartId, bool isBoundaryCondition)
{
    for (size_t i = 0; i < m_partInfoArr.size(); ++i)
    {
        if (m_partInfoArr[i].id == geometryGlobalPartId)
        {
            m_partInfoArr[i].boundaryCondition = isBoundaryCondition;
            return;
        }
    }
    CDPHL_FAIL_MSG("geometryGlobalPartId not found");
}


///--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ElementGroupInfo::setParentPart(cdp_int_t geometryGlobalPartId, cdp_int_t geometryGlobalParentPartId)
{
    for (size_t i = 0; i < m_partInfoArr.size(); ++i)
    {
        if (m_partInfoArr[i].id == geometryGlobalPartId)
        {
            m_partInfoArr[i].parentId = geometryGlobalParentPartId;
            return;
        }
    }
    CDPHL_FAIL_MSG("geometryGlobalPartId not found");
}

///--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
const CDPString& ElementGroupInfo::defaultPartName() const
{
    return m_defaultPartName;
}

///--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
int ElementGroupInfo::defaultPartId() const
{
    return m_defaultPartId;
}


///--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
size_t ElementGroupInfo::partInfoCount() const
{
    return m_partInfoArr.size();
}

///--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
const cdphl::PartInfo& ElementGroupInfo::partInfo(size_t index) const
{
    return m_partInfoArr[index];
}



//==================================================================================================
//
//
//
//==================================================================================================
    
///--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
GeometryInfo::GeometryInfo()
:   m_hasAdaptiveGeometry(false)
{
}

///--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
GeometryInfo::~GeometryInfo()
{
    for (size_t i = 0; i < m_elementGroupInfoArr.size(); ++i)
    {
        delete m_elementGroupInfoArr[i];
    }
}

///--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void GeometryInfo::addElementGroupInfo(cdp_int_t partId, const CDPString& name)
{
    ElementGroupInfo* groupInfo = new ElementGroupInfo(partId, name);
    m_elementGroupInfoArr.push_back(groupInfo);
}
    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
CDPElementGroupInfo* GeometryInfo::addElementGroupInfo()
{
    ElementGroupInfo* groupInfo = new ElementGroupInfo(-1, CDPString());
    m_elementGroupInfoArr.push_back(groupInfo);

    return groupInfo;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void GeometryInfo::setHasAdaptiveGeometry(bool hasAdaptiveGeo)
{
    m_hasAdaptiveGeometry = hasAdaptiveGeo;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
size_t GeometryInfo::elementGroupInfoCount() const
{
    return m_elementGroupInfoArr.size();
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
const ElementGroupInfo& GeometryInfo::elementGroupInfo(size_t index) const
{
    return *m_elementGroupInfoArr[index];
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
bool GeometryInfo::hasAdaptiveGeometry() const
{
    return m_hasAdaptiveGeometry;
}



//==================================================================================================
//
//
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
ResultInfo::ResultInfo(CDPResultType resultType, cdp_int_t id, const CDPString& name, CDPResultMappingType mapping)
:   m_resultType(resultType),
    m_id(id),
    m_name(name),
    m_mapping(mapping),
    m_displacementsContainAbsoluteNodePositions(false),
    m_createDerivedResults(true)
{
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ResultInfo::setIdString(const CDPString &idString)
{
    m_idString = idString;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ResultInfo::setUnit(const CDPString& unit)
{
    m_unit = unit;
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ResultInfo::setDisplacementsContainAbsoluteNodePositions(bool containsAbsoluteNodePositions)
{
    m_displacementsContainAbsoluteNodePositions = containsAbsoluteNodePositions;
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
bool ResultInfo::displacementsContainAbsoluteNodePositions() const
{
    return m_displacementsContainAbsoluteNodePositions;
}



//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ResultInfo::setCreateDerivedResults(bool createDerivedResults)
{
    m_createDerivedResults = createDerivedResults;
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
bool ResultInfo::createDerivedResults() const
{
    return m_createDerivedResults;
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
CDPResultType ResultInfo::type() const
{
    return m_resultType;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
cdp_int_t ResultInfo::id() const
{
    return m_id;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
const CDPString& ResultInfo::name() const
{
    return m_name;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
CDPResultMappingType ResultInfo::mapping() const
{
    return m_mapping;
}

    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
const CDPString& ResultInfo::idString() const
{
    return m_idString;
}

        
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
const CDPString& ResultInfo::unit() const
{
    return m_unit;
}

//==================================================================================================
//
//
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
StateInfo::StateInfo(cdp_int_t id, const CDPString& name, cdp_double_t referenceValue)
:   m_id(id),
    m_name(name),
    m_referenceValue(referenceValue),
    m_groupId(-1)
{
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void StateInfo::setParentStateGroupingId(cdp_int_t groupId)
{
    m_groupId = groupId;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
cdp_int_t StateInfo::id() const
{
    return m_id;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
const CDPString& StateInfo::name() const
{
    return m_name;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
cdp_double_t StateInfo::referenceValue() const
{
    return m_referenceValue;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
cdp_int_t StateInfo::groupId() const
{
    return m_groupId;
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void StateInfo::setName(const CDPString& name)
{
    m_name = name;
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void StateInfo::setReferenceValue(cdp_double_t refVal)
{
    m_referenceValue = refVal;
}


//==================================================================================================
//
//
//
//==================================================================================================
    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
MetaData::MetaData()
:   m_hasRigidBodyTransformations(false),
    m_hasVisibilityResults(false)
{
}
    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
MetaData::~MetaData()
{
    for (size_t i = 0; i < m_geoInfoArr.size(); ++i)
    {
        delete m_geoInfoArr[i];
    }
    
    for (size_t i = 0; i < m_stateInfoArr.size(); ++i)
    {
        delete m_stateInfoArr[i];
    }
    
    for (size_t i = 0; i < m_resultInfoArr.size(); ++i)
    {
        delete m_resultInfoArr[i];
    }
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void MetaData::setSimulationInfo(const CDPString& title, const CDPString& description, CDPSimulationType simulationType, CDPSolutionType solutionType)
{
    m_simulationInfo.title = title;
    m_simulationInfo.description = description;
    m_simulationInfo.simulationType = simulationType;
    m_simulationInfo.solutionType = solutionType;
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
CDPGeometryInfo* MetaData::addGeometryInfo()
{
    GeometryInfo* geoInfo = new GeometryInfo;
    m_geoInfoArr.push_back(geoInfo);

    return geoInfo;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
CDPStateInfo* MetaData::addStateInfo(cdp_int_t stateId, const CDPString& name, cdp_double_t referenceValue)
{
    if (m_stateIdsSet.find(stateId) != m_stateIdsSet.end())
    {
        deleteStateInfo(stateId);
    }

    StateInfo* stateInfo = new StateInfo(stateId, name, referenceValue);
    m_stateInfoArr.push_back(stateInfo);
    m_stateIdsSet.insert(stateId);
    
    return stateInfo;
}
    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
bool MetaData::addStateGroupingInfo(cdp_int_t groupId, const CDPString& name)
{
    if (m_stateGroupIdsSet.find(groupId) != m_stateGroupIdsSet.end())
    {
        return false;
    }

    StateGroupInfo info(groupId, name);
    m_stateGroupInfoArr.push_back(info);
    m_stateGroupIdsSet.insert(groupId);

    return true;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
bool MetaData::addSetInfo(CDPSetType setType, cdp_int_t setId, const CDPString& name)
{
    if (m_setIdsSet.find(setId) != m_setIdsSet.end())
    {
        return false;
    }

    SetInfo info(setType, setId, name);
    m_setInfoArr.push_back(info);
    m_setIdsSet.insert(setId);

    return true;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
CDPResultInfo* MetaData::addResultInfo(CDPResultType resultType, cdp_int_t resultId, const CDPString& name, CDPResultMappingType mapping)
{
    if (m_resultIdsSet.find(resultId) != m_resultIdsSet.end())
    {
        return NULL;
    }

    ResultInfo* resInfo = new ResultInfo(resultType, resultId, name, mapping);
    m_resultInfoArr.push_back(resInfo);
    m_resultIdsSet.insert(resultId);

    return resInfo;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void MetaData::setHasRigidBodyTransformations(bool hasTransformations)
{
    m_hasRigidBodyTransformations = hasTransformations;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void MetaData::setHasElementVisibilityResults(bool hasVisibilityResults)
{
    m_hasVisibilityResults = hasVisibilityResults;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
size_t MetaData::stateInfoCount() const
{
    return m_stateInfoArr.size();
}
    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
const StateInfo& MetaData::stateInfo(size_t index) const
{
    CDPHL_ASSERT(m_stateInfoArr[index]);
    return *m_stateInfoArr[index];
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
const std::set<cdp_int_t>& MetaData::stateIdsSet() const
{
    return m_stateIdsSet;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
size_t MetaData::stateGroupInfoCount() const
{
    return m_stateGroupInfoArr.size();
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
const StateGroupInfo& MetaData::stateGroupInfo(size_t index) const
{
    return m_stateGroupInfoArr[index];
}
    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
size_t MetaData::setInfoCount() const
{
    return m_setInfoArr.size();
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
const cdphl::SetInfo& MetaData::setInfo(size_t index) const
{
    return m_setInfoArr[index];
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
size_t MetaData::resultInfoCount() const
{
    return m_resultInfoArr.size();
}
    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
const ResultInfo& MetaData::resultInfo(size_t index) const
{
    CDPHL_ASSERT(m_resultInfoArr[index]);
    return *m_resultInfoArr[index];
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
bool MetaData::hasRigidBodyTransformations() const
{
    return m_hasRigidBodyTransformations;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
bool MetaData::hasVisibilityResults() const
{
    return m_hasVisibilityResults;;
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
SimulationInfo MetaData::simulationInfo() const
{
    return m_simulationInfo;
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
size_t MetaData::geometryInfoCount() const
{
    return m_geoInfoArr.size();
}
    

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
const GeometryInfo& MetaData::geometryInfo(size_t index) const
{
    CDPHL_ASSERT(m_geoInfoArr[index]);
    return *m_geoInfoArr[index];
}
    

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void MetaData::deleteResultInfo(cdp_int_t resultId)
{
    for (size_t i = 0; i < m_resultInfoArr.size(); ++i)
    {
        if (m_resultInfoArr[i]->id() == resultId)
        {
            delete m_resultInfoArr[i];
            m_resultInfoArr.erase(m_resultInfoArr.begin() + static_cast<ptrdiff_t>(i));
            return;
        }
    }
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void MetaData::deleteStateInfo(cdp_int_t stateId)
{
    for (size_t i = 0; i < m_stateInfoArr.size(); ++i)
    {
        if (m_stateInfoArr[i]->id() == stateId)
        {
            delete m_stateInfoArr[i];
            m_stateInfoArr.erase(m_stateInfoArr.begin() + static_cast<ptrdiff_t>(i));
            return;
        }
    }
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void MetaData::changeStateInfo(cdp_int_t stateId, const CDPString& name, cdp_double_t referenceValue)
{
    for (size_t i = 0; i < m_stateInfoArr.size(); ++i)
    {
        if (m_stateInfoArr[i]->id() == stateId)
        {
            m_stateInfoArr[i]->setName(name);
            m_stateInfoArr[i]->setReferenceValue(referenceValue);
            
            return;
        }
    }
}

}  // cdphl
