//##################################################################################################
//
//  Ceetron Data Provider Framework
//
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2018, Ceetron AS
//##################################################################################################

#pragma once

#include "cdphlBase.h"

#include "CDPMetaData.h"
#include "CDPString.h"

#include <vector>
#include <set>
#include <map>
#include <string>

namespace cdphl {


//==================================================================================================
//
//
//
//==================================================================================================
struct SimulationInfo
{
    CDPString title;
    CDPString description;
    CDPSimulationType simulationType;
    CDPSolutionType solutionType;
    
    SimulationInfo();
};


//==================================================================================================
//
// 
//
//==================================================================================================
class StateInfo : public CDPStateInfo
{
public:
    StateInfo(cdp_int_t id, const CDPString& name, cdp_double_t referenceValue);

    virtual void setParentStateGroupingId(cdp_int_t groupId);
    
    cdp_int_t           id() const;
    const CDPString&    name() const;
    cdp_double_t        referenceValue() const;
    cdp_int_t           groupId() const;
    
    void                setName(const CDPString& name);
    void                setReferenceValue(cdp_double_t refVal);
    
private:
    cdp_int_t       m_id;
    CDPString       m_name;
    cdp_double_t    m_referenceValue;
    cdp_int_t       m_groupId;
};

    
//==================================================================================================
//
//
//
//==================================================================================================
struct StateGroupInfo
{
    cdp_int_t     id;
    CDPString     name;
    
    StateGroupInfo(cdp_int_t inId, const CDPString& inName)
    :   id(inId),
        name(inName)
    {}
};

//==================================================================================================
//
// 
//
//==================================================================================================
struct PartInfo
{
    cdp_int_t       id;
    CDPString       name;
    bool            boundaryCondition;

    cdp_int_t       parentId;

    std::map<std::string, std::string> partData;

    PartInfo(cdp_int_t inId, const CDPString& inName, bool inBondaryCondition = false, cdp_int_t intParentPartID = -1)
    :   id(inId),
        name(inName),
        boundaryCondition(inBondaryCondition),
        parentId(intParentPartID)
    {}
};


//==================================================================================================
//
//
//
//==================================================================================================
class ElementGroupInfo : public CDPElementGroupInfo
{
public:
    ElementGroupInfo(cdp_int_t defaultPartId, const CDPString& defaultPartName);

    virtual void        setPartName(cdp_int_t geometryGlobalPartId, const CDPString& partName) override;
    virtual void        setPartBoundaryCondition(cdp_int_t geometryGlobalPartId, bool isBoundaryCondition) override;
    virtual void        setParentPart(cdp_int_t geometryGlobalPartId, cdp_int_t geometryGlobalParentPartId) override;

    virtual void        addPartData(cdp_int_t geometryGlobalPartId, const CDPString& key, const CDPString& value) override;
    
    const CDPString&    defaultPartName() const;
    int                 defaultPartId() const;

    size_t              partInfoCount() const;
    const PartInfo&     partInfo(size_t index) const;

private:
    cdp_int_t               m_defaultPartId;
    CDPString               m_defaultPartName;
    std::vector<PartInfo>   m_partInfoArr;
};


//==================================================================================================
//
// 
//
//==================================================================================================
class GeometryInfo : public CDPGeometryInfo
{
public:
    GeometryInfo();
    ~GeometryInfo();
    
    virtual void                    addElementGroupInfo(cdp_int_t partId, const CDPString& name) override;
    virtual CDPElementGroupInfo*    addElementGroupInfo() override;
    
    virtual void                    setHasAdaptiveGeometry(bool hasAdaptiveGeo) override;

    size_t                          elementGroupInfoCount() const;
    const ElementGroupInfo&         elementGroupInfo(size_t index) const;

    bool                            hasAdaptiveGeometry() const;

private:
    std::vector<ElementGroupInfo*>  m_elementGroupInfoArr;
    bool                            m_hasAdaptiveGeometry;
};
    

//==================================================================================================
//
// 
//
//==================================================================================================
class ResultInfo : public CDPResultInfo
{
public:
    ResultInfo(CDPResultType resultType, cdp_int_t id, const CDPString& name, CDPResultMappingType mapping);

    virtual void            setUnit(const CDPString& unit) override;
    virtual void            setIdString(const CDPString& idString) override;

    virtual void            setDisplacementsContainAbsoluteNodePositions(bool containsAbsoluteNodePositions) override;
    virtual void            setCreateDerivedResults(bool createDerivedResults) override;

    CDPResultType           type() const;
    cdp_int_t               id() const;
    const CDPString&        name() const;
    CDPResultMappingType    mapping() const;
    const CDPString&        idString() const;
    const CDPString&        unit() const;

    bool                    displacementsContainAbsoluteNodePositions() const;
    bool                    createDerivedResults() const;

private:
    CDPResultType           m_resultType;
    cdp_int_t               m_id;
    CDPString               m_name;
    CDPResultMappingType    m_mapping;
    CDPString               m_idString;
    CDPString               m_unit;

    bool                    m_displacementsContainAbsoluteNodePositions;
    bool                    m_createDerivedResults;
};


//==================================================================================================
//
// 
//
//==================================================================================================
struct SetInfo {
    CDPSetType      setType;
    cdp_int_t       id;
    CDPString       name;

    SetInfo(CDPSetType inSetType, cdp_int_t inId, const CDPString& inName)
    :   setType(inSetType),
        id(inId),
        name(inName)
    {}
};


//==================================================================================================
//
// 
//
//==================================================================================================
class MetaData : public CDPMetaData
{
public:
    MetaData();
    ~MetaData();
    
    virtual void                setSimulationInfo(const CDPString& title, const CDPString& description, CDPSimulationType simulationType, CDPSolutionType solutionType) override;
    virtual CDPGeometryInfo*    addGeometryInfo() override;
    virtual CDPStateInfo*       addStateInfo(cdp_int_t stateId, const CDPString& name, cdp_double_t referenceValue) override;
    virtual bool                addStateGroupingInfo(cdp_int_t groupId, const CDPString& name) override;
    virtual bool                addSetInfo(CDPSetType setType, cdp_int_t setId, const CDPString& name) override;
    virtual CDPResultInfo*      addResultInfo(CDPResultType resultType, cdp_int_t resultId, const CDPString& name, CDPResultMappingType mapping) override;
    virtual void                setHasRigidBodyTransformations(bool hasTransformations) override;
    virtual void                setHasElementVisibilityResults(bool hasVisibilityResults) override;

    void                        deleteResultInfo(cdp_int_t resultId);
    void                        deleteStateInfo(cdp_int_t stateId);
    
    void                        changeStateInfo(cdp_int_t stateId, const CDPString& name, cdp_double_t referenceValue);
    
    SimulationInfo              simulationInfo() const;

    size_t                      geometryInfoCount() const;
    const GeometryInfo&         geometryInfo(size_t index) const;
    
    size_t                      stateInfoCount() const;
    const StateInfo&            stateInfo(size_t index) const;
    const std::set<cdp_int_t>&  stateIdsSet() const;

    size_t                      stateGroupInfoCount() const;
    const StateGroupInfo&       stateGroupInfo(size_t index) const;
    
    size_t                      setInfoCount() const;
    const SetInfo&              setInfo(size_t index) const;

    size_t                      resultInfoCount() const;
    const ResultInfo&           resultInfo(size_t index) const;
    bool                        hasRigidBodyTransformations() const;
    bool                        hasVisibilityResults() const;

private:
    SimulationInfo                  m_simulationInfo;
    std::vector<GeometryInfo*>      m_geoInfoArr;
    std::vector<StateInfo*>         m_stateInfoArr;
    std::vector<StateGroupInfo>     m_stateGroupInfoArr;
    std::vector<SetInfo>            m_setInfoArr;
    std::vector<ResultInfo*>        m_resultInfoArr;
    bool                            m_hasRigidBodyTransformations;
    bool                            m_hasVisibilityResults;

    std::set<cdp_int_t>             m_stateIdsSet;
    std::set<cdp_int_t>             m_stateGroupIdsSet;
    std::set<cdp_int_t>             m_setIdsSet;
    std::set<cdp_int_t>             m_resultIdsSet;
};


}  // cdphl

