//##################################################################################################
//
//  Ceetron Data Provider Framework
//
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2018, Ceetron AS
//##################################################################################################

#include "cdphlPluginManager.h"
#include "cdphlDataProviderFactory.h"
#include "cdphlDynamicLibrary.h"
#include "cdphlBase.h"
#include "cdphlDefaultLogger.h"
#include "cdphlStringUtils.h"
#include "cdphlAssert.h"

#include "CDPPlugin.h"
#include "CDPDataProviderFactory.h"


#if defined(__GNUC__) && (CDPHL_GCC_VER >= 40500)
CDPHL_GCC_DIAGNOSTIC_IGNORE("-Wpedantic")
#endif

#ifdef _MSC_VER
#pragma warning (disable: 4640)
#endif


// Helper macros for writing log messages to a logger
#define CDPHL_LOG_ERROR(theLogger, theMessage)    if ((theLogger))      { (theLogger)->log(Logger::LL_ERROR, (theMessage)); }
#define CDPHL_LOG_WARNING(theLogger, theMessage)  if ((theLogger))      { (theLogger)->log(Logger::LL_WARNING, (theMessage)); }
#define CDPHL_LOG_INFO(theLogger, theMessage)     if ((theLogger))      { (theLogger)->log(Logger::LL_INFO, (theMessage)); }
#define CDPHL_LOG_DEBUG(theLogger, theMessage)    if ((theLogger))      { (theLogger)->log(Logger::LL_DEBUG, (theMessage)); }



namespace cdphl {


#if defined(WIN32)
static const std::string DYNLIB_EXTENSION("dll");
#elif defined(__APPLE__)
static const std::string DYNLIB_EXTENSION("dylib");
#else
static const std::string DYNLIB_EXTENSION("so");
#endif


//==================================================================================================
//
// 
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
PluginManager::PluginManager()
:   m_logger(new DefaultLogger),
    m_pluginBeingInitialized(NULL),
    m_lastAssignedSequenceNumber(0)
{
    m_frameworkServices.pluginApiMajorVer = CDP_API_MAJOR_VER;
    m_frameworkServices.pluginApiMinorVer = CDP_API_MINOR_VER;
    m_frameworkServices.registerProviderFunc = staticHandleProviderRegistration;
    m_frameworkServices.logFunc = staticLogFunction;
    m_frameworkServices.setPluginInfoFunc = staticHandleSetPluginInfo;
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
PluginManager::~PluginManager()
{
    CDPHL_LOG_DEBUG(m_logger, "PluginManager::~PluginManager()");

    shutdown();
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
bool PluginManager::hasProviderWithId(const char* providerIdString) const
{
    for (size_t i = 0; i < m_providerFactoryArr.size(); i++)
    {
        if (m_providerFactoryArr[i]->providerId() == providerIdString)
        {
            return true;
        }
    }

    return false;
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
PluginManager* PluginManager::instance()
{
    static PluginManager sl_pluginManager;
    return &sl_pluginManager;
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void PluginManager::shutdown()
{
    CDPHL_LOG_DEBUG(m_logger, "PluginManager::shutdown()\n");

    for (size_t i = 0; i < m_providerFactoryArr.size(); i++)
    {
        delete m_providerFactoryArr[i];
    }

    m_providerFactoryArr.clear();


    for (size_t i = 0; i < m_pluginEntryArr.size(); i++)
    {
        PluginEntry entry = m_pluginEntryArr[i];
        entry.exitFunc();
        delete entry.dynLib;
    }

    m_pluginEntryArr.clear();
}

//--------------------------------------------------------------------------------------------------
/// Note that the logger will be woned and deleted by the plugin manager
//--------------------------------------------------------------------------------------------------
void PluginManager::setLogger(Logger* logger)
{
    if (m_logger != logger)
    {
        delete m_logger;
        m_logger = logger;
    }
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
bool PluginManager::loadPlugin(const std::string& pluginBaseFilename)
{
    std::string fullPluginFilename = pluginBaseFilename + "." + DYNLIB_EXTENSION;

    // Make sure we don't load the same plugin twice
    for (size_t i = 0; i < m_pluginEntryArr.size(); i++)
    {
        if (m_pluginEntryArr[i].fullPluginFilename == fullPluginFilename)
        {
            CDPHL_LOG_ERROR(m_logger, "The plugin '" + fullPluginFilename + "' is already loaded");
            return false;
        }
    }

    std::string err;
    DynamicLibrary* dynLib = DynamicLibrary::load(fullPluginFilename.c_str(), &err);
    if (!dynLib)
    {
        CDPHL_LOG_ERROR(m_logger, "Error loading plugin dynamic library: " + err);
        return false;
    }

    // Get the plugin init function
    CDP_PluginInitFunc pluginInitFunc = reinterpret_cast<CDP_PluginInitFunc>(dynLib->getSymbol("initializeDataProviderPlugin"));
    if (!pluginInitFunc)
    {
        // Dynamic library missing entry point?
        CDPHL_LOG_ERROR(m_logger, "Could not find initializeDataProviderPlugin entry point in plugin");
        delete dynLib;
        return false;
    }

    if (!doInitializationOfPlugin(pluginInitFunc, dynLib, fullPluginFilename))
    {
        CDPHL_LOG_ERROR(m_logger, "Error initializing dynamic plugin");
        delete dynLib;
        return false;
    }

    return true;
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
bool PluginManager::doInitializationOfPlugin(CDP_PluginInitFunc pluginInitFunc, DynamicLibrary* dynLib, std::string fullPluginFilename)
{
    PluginEntry pluginEntry;
    pluginEntry.fullPluginFilename = fullPluginFilename;
    pluginEntry.dynLib = dynLib;

    m_pluginBeingInitialized = &pluginEntry;
    CDP_PluginExitFunc pluginExitFunc = pluginInitFunc(&m_frameworkServices);
    m_pluginBeingInitialized = NULL;

    if (!pluginExitFunc)
    {
        return false;
    }

    pluginEntry.exitFunc = pluginExitFunc;
    m_pluginEntryArr.push_back(pluginEntry);

    CDPHL_LOG_DEBUG(m_logger, "Plugin successfully initialized");
    CDPHL_LOG_DEBUG(m_logger, "  pluginFileName: " + fullPluginFilename);
    
    const size_t numInfoKeys = pluginEntry.pluginInfoKeyValues.size();
    CDPHL_LOG_DEBUG(m_logger, "  pluginInfo(" + StringUtils::toStdString(numInfoKeys) + ")");

    for (std::map<std::string, std::string>::iterator it = pluginEntry.pluginInfoKeyValues.begin(); it != pluginEntry.pluginInfoKeyValues.end(); ++it)
    {
        CDPHL_LOG_DEBUG(m_logger, "    " + it->first + ": " + it->second);
    }

    return true;
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
bool PluginManager::initializeStaticPlugin(CDP_PluginInitFunc pluginInitFunc)
{
    CDPHL_ASSERT(pluginInitFunc);

    if (!doInitializationOfPlugin(pluginInitFunc, NULL, std::string()))
    {
        CDPHL_LOG_ERROR(m_logger, "Error initializing static plugin");
        return false;
    }

    return true;
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
size_t PluginManager::providerCount() const
{
    return m_providerFactoryArr.size();
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
DataProviderFactory* PluginManager::providerFactory(size_t index) const
{
    return m_providerFactoryArr[index];
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
int PluginManager::handleProviderRegistration(const char* providerIdString, const CDP_RegisterProviderParams* params)
{
    CDPHL_LOG_DEBUG(m_logger, "PluginManager::handleProviderRegistration()");
    CDPHL_LOG_DEBUG(m_logger, "  providerId: " + StringUtils::toStdString(providerIdString));
    CDPHL_LOG_DEBUG(m_logger, "  pluginApiVersion: " + StringUtils::toStdString(params->pluginApiMajorVer) + ", " + StringUtils::toStdString(params->pluginApiMinorVer));

    if (params->pluginApiMajorVer != CDP_API_MAJOR_VER)
    {
        CDPHL_LOG_ERROR(m_logger, "The plugin API Major Version (" + StringUtils::toStdString(params->pluginApiMajorVer) + ") does not match the host's version (" + StringUtils::toStdString(CDP_API_MAJOR_VER) + ").");
        params->destroyProviderFactoryFunc(params->providerFactoryObj);
        return CDP_RET_FAIL;
    }
    
    // Accept plugins with MINOR version less or equal to the host (C3 / AD)
    //
    // E.g.: 3.1 C3 server accepts 3.0 (or 3.1) plugins
    //       3.1 C3 server does NOT accept 3.2 plugins
    //
    // So the rule is:
    //  - If adding/changing a call from HOST -> Plugin, increase MAJOR version
    //  - If adding a call from Plugin -> Host, increase MINOR version
    if (params->pluginApiMinorVer > CDP_API_MINOR_VER)
    {
        CDPHL_LOG_ERROR(m_logger, "The plugin API Minor Version (" + StringUtils::toStdString(params->pluginApiMinorVer) + ") is not compatible with host's version (" + StringUtils::toStdString(CDP_API_MINOR_VER) + ").");
        params->destroyProviderFactoryFunc(params->providerFactoryObj);
        return CDP_RET_FAIL;
    }

    if (hasProviderWithId(providerIdString))
    {
        CDPHL_LOG_ERROR(m_logger, "A provider with idString '" + StringUtils::toStdString(providerIdString) + "' is already present");
        params->destroyProviderFactoryFunc(params->providerFactoryObj);
        return CDP_RET_FAIL;
    }

    CDPDataProviderFactory* cdpFactory = params->providerFactoryObj;
    if (!cdpFactory)
    {
        return CDP_RET_FAIL;
    }

    const int sequenceNumber = ++m_lastAssignedSequenceNumber;
    DataProviderFactory* factory = new DataProviderFactory(sequenceNumber, providerIdString, cdpFactory, params->destroyProviderFactoryFunc);
    m_providerFactoryArr.push_back(factory);

    return CDP_RET_SUCCESS;
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
int PluginManager::staticHandleProviderRegistration(const char* providerIdString, const CDP_RegisterProviderParams* params)
{
    PluginManager* mgr = PluginManager::instance();
    return mgr->handleProviderRegistration(providerIdString, params);
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void PluginManager::handleSetPluginInfo(const char* key, const char* value)
{
    if (!m_pluginBeingInitialized)
    {
        CDPHL_LOG_ERROR(m_logger, "Cannot set plugin info outside of plugin initialization");
        return;
    }

    if (!key || !value)
    {
        CDPHL_LOG_ERROR(m_logger, "Cannot set plugin info, either key or value is invalid");
        return;
    }

    m_pluginBeingInitialized->pluginInfoKeyValues[std::string(key)] = value;
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void PluginManager::staticHandleSetPluginInfo(const char* key, const char* value)
{
    PluginManager* mgr = PluginManager::instance();
    return mgr->handleSetPluginInfo(key, value);
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void PluginManager::staticLogFunction(int level, const char* message)
{
    PluginManager* mgr = PluginManager::instance();
    if (mgr->m_logger)
    {
        Logger::Level loggerLevel = Logger::LL_ERROR;
        if      (level == 1) loggerLevel = Logger::LL_ERROR;
        else if (level == 2) loggerLevel = Logger::LL_WARNING;
        else if (level == 3) loggerLevel = Logger::LL_INFO;
        else if (level == 4) loggerLevel = Logger::LL_DEBUG;

        mgr->m_logger->log(loggerLevel, message);
    }
}


} // cdphl

