//##################################################################################################
//
//  Ceetron Data Provider Framework
//
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2018, Ceetron AS
//##################################################################################################

#include "cdphlValidator.h"
#include "cdphlElementGroup.h"
#include "cdphlNodeGroup.h"
#include "cdphlInitOptions.h"

#include <set>
#include <cstdarg>
#include <cstdio>

namespace cdphl {

//==================================================================================================
//
// 
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
Validator::Validator(Logger* logger)
: m_logger(logger)
{
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
bool Validator::validate(CDPDataProvider* provider, std::string modelKey)
{
    const bool initOk = provider->init(CDPString::fromChar(modelKey.c_str()), InitOptions());
    if (!initOk)
    {
        logError("Failed to initialize with modelKey: %s", modelKey.c_str());
        provider->close();
        return false;
    }

    cdphl::MetaData metaData;
    provider->getMetaData(&metaData);
    
    
    std::vector<cdp_int_t> allStateIds = getAllStateIds(metaData);
    if (allStateIds.size() == 0)
    {
        logError("No states found!");
        return false;
    }
    
    if (!validateMetaData(metaData))
    {
        logError("Failed to validate meta data");
        return false;
    }
    
    const size_t geoCount = metaData.geometryInfoCount();
    for (size_t ig = 0; ig < geoCount; ig++)
    {
        const cdp_size_t geometryIndex = ig;
        const cdphl::GeometryInfo& geoInfo = metaData.geometryInfo(ig);
        
        if (geoInfo.hasAdaptiveGeometry())
        {
            for (size_t is = 0; is < allStateIds.size(); is++)
            {
                const cdp_int_t stateId = allStateIds[is];
                cdphl::GeometrySpec spec(geometryIndex, stateId);
                cdphl::Geometry geo;
                const bool geoOk = provider->getGeometry(spec, &geo);
                if (!geoOk)
                {
                    logError("Failed to get adaptive geometry with geometryIndex=%d stateId=%d", geometryIndex, stateId);
                    return false;
                }
                
                // Need to do this to complete the element definitions
                geo.computeDerivedData();
                
                if (!validateGeometry(geo, geoInfo))
                {
                    logError("Failed to validate geometry with geometryIndex=%d stateId=%d", geometryIndex, stateId);
                    return false;
                }
            }
        }
        else
        {
            const cdp_int_t firstStateId = allStateIds[0];
            cdphl::GeometrySpec spec(geometryIndex, firstStateId);
            cdphl::Geometry geo;
            const bool geoOk = provider->getGeometry(spec, &geo);
            if (!geoOk)
            {
                logError("Failed to get fixed geometry with geometryIndex=%d stateId=%d", geometryIndex, firstStateId);
                continue;
            }
            
            // Need to do this to complete the element definitions
            geo.computeDerivedData();
            
            if (!validateGeometry(geo, geoInfo))
            {
                logError("Failed to validate geometry with geometryIndex=%d stateId=%d", geometryIndex, firstStateId);
                return false;
            }
        }
    }

    return true;
}
    

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
bool Validator::validateMetaData(const cdphl::MetaData& metaData)
{
    const size_t geoCount = metaData.geometryInfoCount();
    for (size_t geoIdx = 0; geoIdx < geoCount; ++geoIdx)
    {
        const cdphl::GeometryInfo& geoInfo = metaData.geometryInfo(geoIdx);
        
        std::set<int> partIdArr;
        
        const size_t eltGroupCount = geoInfo.elementGroupInfoCount();
        for (size_t grpIdx = 0; grpIdx < eltGroupCount; ++grpIdx)
        {
            const cdphl::ElementGroupInfo& eltGrpInfo = geoInfo.elementGroupInfo(grpIdx);
            
            if (eltGrpInfo.defaultPartId() >= 0)
            {
                int partId = eltGrpInfo.defaultPartId();

                if (partIdArr.find(partId) != partIdArr.end())
                {
                    logError("ElementGroup[idx=%d] has defaultPartId=%d which is not unique for the geometry.", grpIdx, partId);
                    return false;
                }
                
                partIdArr.insert(partId);
            }
            
            for (size_t partIdx = 0; partIdx < eltGrpInfo.partInfoCount(); ++partIdx)
            {
                int partId = eltGrpInfo.partInfo(partIdx).id;
                
                if (partIdArr.find(partId) != partIdArr.end())
                {
                    logError("ElementGroup[idx=%d] has PartInfo with id=%d which is not unique for the geometry.", grpIdx, partId);
                    return false;
                }
                
                partIdArr.insert(partId);
            }
        }
    }

    return true;
}

    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
bool Validator::validateGeometry(const cdphl::Geometry& geo, const cdphl::GeometryInfo& geoInfo)
{
    if (geo.elementGroupCount() != geoInfo.elementGroupInfoCount())
    {
        logError("Meta data does not match geometry data. elementGroupInfoCount(%d) != elementGroupCount(%d).", geoInfo.elementGroupInfoCount(), geo.elementGroupCount());
        return false;
    }

    const size_t eltGroupCount = geo.elementGroupCount();
    for (size_t igrp = 0; igrp < eltGroupCount; igrp++)
    {
        const cdphl::ElementGroup& eltGrp = geo.elementGroup(igrp);
        const cdphl::ElementGroupInfo& eltGrpInfo = geoInfo.elementGroupInfo(igrp);
        
        const size_t nodeGrpIdx = eltGrp.nodeGroupIndex();
        
        if (nodeGrpIdx >= geo.nodeGroupCount())
        {
            logError("ElementGroup[idx=%d] refers to NodeGroup[idx=%d] which does not exist.", igrp, nodeGrpIdx);
            return false;
        }
        
        // Validate part ids
        if (eltGrp.elementPartIdArr().size() > 0)
        {
            std::set<int> partIdArr;
            for (size_t pi = 0; pi < eltGrpInfo.partInfoCount(); ++pi)
            {
                const cdphl::PartInfo& partInfo = eltGrpInfo.partInfo(pi);
                if (partIdArr.find(partInfo.id) != partIdArr.end())
                {
                    logError("ElementGroup[idx=%d]: Duplicate part ids found in elementGroupInfo. id=%d", igrp, partInfo.id);
                    return false;
                }
                
                partIdArr.insert(partInfo.id);
            }
            
            if (eltGrp.elementPartIdArr().size() != eltGrp.elementCount())
            {
                logError("ElementGroup[idx=%d]: Number of perElementPartIDs specified does not match number of elements.");
                return false;
            }
            
            for (size_t eltIdx = 0; eltIdx < eltGrp.elementPartIdArr().size(); ++eltIdx)
            {
                int eltPartId = eltGrp.elementPartIdArr()[eltIdx];
                if (partIdArr.find(eltPartId) == partIdArr.end())
                {
                    logError("ElementGroup[idx=%d]: Element[idx=%d] refers partId=%d which is not defined in ElementGroupInfo::partInfo.", igrp, eltIdx, eltPartId);
                    return false;
                }
            }
        }
        
        const size_t nodeCount = geo.nodeGroup(nodeGrpIdx).nodeArr().size()/3;
        
        const size_t subGroupDataCount = eltGrp.subGroupDataCount();
        for (size_t isub = 0; isub < subGroupDataCount; isub++)
        {
            const cdphl::ElementSubGroupData& subGroup = eltGrp.subGroupData(isub);
            
            // Validate element nodes
            for (size_t eltNodeIdx = 0; eltNodeIdx < subGroup.elementNodeArr.size(); ++eltNodeIdx)
            {
                int nodeIdx = subGroup.elementNodeArr[eltNodeIdx];
                if (nodeIdx < 0 || nodeIdx >= static_cast<int>(nodeCount))
                {
                    logError("ElementGroup[idx=%d] refers to node with index %d (in NodeGroup[idx=%d]) which does not exist (max idx: %d)", igrp, nodeIdx, nodeGrpIdx, nodeCount - 1);
                    return false;
                }
            }
            
            for (size_t eltIdx = 0; eltIdx < subGroup.elementCount; ++eltIdx)
            {
                CDPElementType eltType = subGroup.elementTypeArr.size() > 0 ? subGroup.elementTypeArr[eltIdx] : subGroup.elementType;
                
                if (eltType != CDP_ELT_POLYHEDRONS)
                {
                    int actualNumEltNodes = subGroup.numElementNodes(eltIdx);
                    int expectedNumEltNodes = static_cast<int>(ElementGroup::fixedNodeCountForElement(eltType));
                    
                    if (actualNumEltNodes != expectedNumEltNodes)
                    {
                        logError("ElementGroup idx=%d (with NodeGroup idx=%d) does not have the correct amount of element nodes for element idx=%d. Expected: %d, Actual: %d.", igrp, nodeGrpIdx, eltIdx, expectedNumEltNodes, actualNumEltNodes);
                        return false;
                    }
                }
            }
        }
    }

    return true;
}
    
    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void Validator::logError(const char* format, ...)
{
    va_list argList;
    va_start(argList, format);
    
    const int maxFormatLength = 4000;
    char temp[maxFormatLength + 1];
    
#ifdef WIN32
    _vsnprintf_s(temp, maxFormatLength, format, argList);
#else
    vsprintf(temp, format, argList);
#endif
    
    va_end(argList);
    
    m_logger->log(Logger::LL_ERROR, temp);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
std::vector<cdp_int_t> Validator::getAllStateIds(const cdphl::MetaData& metaData)
{
    std::vector<cdp_int_t> allIds;
    
    const size_t stateCount = metaData.stateInfoCount();
    for (size_t i = 0; i < stateCount; i++)
    {
        const cdphl::StateInfo& stateInfo = metaData.stateInfo(i);
        allIds.push_back(stateInfo.id());
    }
    
    return allIds;
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
std::vector<cdp_int_t> Validator::getAllResultIds(const cdphl::MetaData& metaData)
{
    std::vector<cdp_int_t> allIds;
    
    const size_t resultCount = metaData.resultInfoCount();
    for (size_t i = 0; i < resultCount; i++)
    {
        const cdphl::ResultInfo& resInfo = metaData.resultInfo(i);
        allIds.push_back(resInfo.id());
    }
    
    return allIds;
}


} // cdphl
