//##################################################################################################
//
//  Ceetron Data Provider Framework
//
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2018, Ceetron AS
//##################################################################################################

#pragma once

#include "CDPBaseTypes.h"
#include "CDPVersion.h"

class CDPDataProviderFactory;

#define CDP_RET_SUCCESS 1
#define CDP_RET_FAIL    0

extern "C" {


/// Type definition of function pointer for deleting instances of CDPDataProviderFactory
typedef void (*CDP_DestroyProviderFactoryFunc)(CDPDataProviderFactory* objPtr);

//==================================================================================================
///
/// \struct CDP_RegisterProviderParams
/// 
/// Parameter struct utilized when registering data providers
//==================================================================================================
struct CDP_RegisterProviderParams
{
    cdp_int_t                       pluginApiMajorVer;          ///< Major version of the framework plugin API for which this plugin was written
    cdp_int_t                       pluginApiMinorVer;          ///< Minor version of the framework plugin API for which this plugin was written
    CDPDataProviderFactory*         providerFactoryObj;         ///< Pointer to CDPDataProviderFactory object that allows creation/destruction of CDPDataProvider instances
    CDP_DestroyProviderFactoryFunc  destroyProviderFactoryFunc; ///< Pointer to function that should be used to delete providerFactoryObj
};

/// Type definition of the function pointer for registering data providers with the plugin manager
typedef int (*CDP_RegisterProviderFunc)(const char* providerIdString, const CDP_RegisterProviderParams* params);



/// Type definition of log function provided by the framework
/// Log levels are 1=error, 2=warning, 3=info, 4=debug
typedef void (*CDP_LogFunc)(int level, const char* message);


/// Type definition of function provided by the framework for setting plugin information in the form of key, value pairs
typedef void (*CDP_SetPluginInfoFunc)(const char* key, const char* value);


//==================================================================================================
///
/// \struct CDP_FrameworkServices
/// 
/// Aggregates the services that the plugin framework provides to the plugin.
/// This includes version of the plugin API, registering data providers and functions for logging and reporting information on the plugin itself. 
/// This struct is passed to each plugin at initialization time.
//==================================================================================================
struct CDP_FrameworkServices
{
    cdp_int_t                   pluginApiMajorVer;      ///< Major version of the framework plugin API
    cdp_int_t                   pluginApiMinorVer;      ///< Minor version of the framework plugin API
    CDP_RegisterProviderFunc    registerProviderFunc;   ///< Function for registering data providers
    CDP_LogFunc                 logFunc;                ///< Function pointer for log function
    CDP_SetPluginInfoFunc       setPluginInfoFunc;      ///< Function pointer for communicating plugin info
};


// Type definition of the plugin exit function pointer (used by the plugin manager to tell plugins to cleanup before being unloaded)
typedef void (*CDP_PluginExitFunc)();


/// Type definition definition of the plugin initialization function pointer (used by plugin manager to initialize plugins)
/// Note that the return type is the CDP_PluginExitFunc (used by plugin manager to tell plugins to cleanup). 
/// If initialization failed for any reason, it must return NULL to let the plugin manger know that the plugin 
/// wasn't initialized properly. 
/// The plugin should use the framework services to register the data providers it exports as well as communicating log messages.
typedef CDP_PluginExitFunc (*CDP_PluginInitFunc)(const CDP_FrameworkServices* frameworkServices);



// Make sure that CDP_BUILD_PLUGIN_DLL is defined when building the plugin DLL.
#ifdef WIN32
    #ifdef CDP_BUILD_PLUGIN_DLL
        #define CDP_PLUGIN_EXPORT __declspec(dllexport)
    #else
        #define CDP_PLUGIN_EXPORT __declspec(dllimport)
    #endif
#else
    #define CDP_PLUGIN_EXPORT
#endif


/// Named exported entry point into the plugin
/// This definition is required even though the function is loaded from a dynamic library by name and cast to CDP_PluginInitFunc
/// If this declaration is commented out DynamicLibrary::getSymbol() fails
/// The plugin's initialization function MUST be called "initializeDataProviderPlugin" and conform to the CDP_PluginInitFunc signature.
extern "C" CDP_PLUGIN_EXPORT CDP_PluginExitFunc initializeDataProviderPlugin(const CDP_FrameworkServices* params);

} // extern "C"
