//##################################################################################################
//
//  Ceetron Data Provider Framework
//
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2018, Ceetron AS
//##################################################################################################

//==================================================================================================
///
/// \class CDPString
/// 
/// A simple class for handling strings.
/// 
/// CDP uses UTF-8 based strings. This class stores strings in UTF-8 format, and provides
/// creation from ASCII char and UTF-8 strings.
/// 
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// Constructor
//--------------------------------------------------------------------------------------------------
CDPString::CDPString()
{
    m_str = new cdp_char_t[1];
    m_str[0] = 0;
}

//--------------------------------------------------------------------------------------------------
/// Copy constructor
//--------------------------------------------------------------------------------------------------
CDPString::CDPString(const CDPString& other)
{
    if (other.m_str)
    {
        size_t length = other.byteCount();
        m_str = new cdp_char_t[length + 1];
        memcpy(m_str, other.m_str, length);
        m_str[length] = 0;
    }
    else
    {
        m_str = new cdp_char_t[1];
        m_str[0] = 0;
    }
}

//--------------------------------------------------------------------------------------------------
/// Destructor
//--------------------------------------------------------------------------------------------------
CDPString::~CDPString()
{
    delete[] m_str;
    m_str = NULL;
}

//--------------------------------------------------------------------------------------------------
/// Blank string
//--------------------------------------------------------------------------------------------------
const CDPString& CDPString::emptyString()
{
    static CDPString sm_blank;
    return sm_blank;
}

//--------------------------------------------------------------------------------------------------
/// Assignment operator
//--------------------------------------------------------------------------------------------------
CDPString& CDPString::operator=(const CDPString& rhs)
{
    cdp_char_t* oldStr = m_str;

    const size_t newByteLength = rhs.byteCount();
    m_str = new cdp_char_t[newByteLength + 1];

    if (newByteLength > 0)
    {
        memcpy(m_str, rhs.m_str, newByteLength);
    }
    m_str[newByteLength] = 0;

    delete[] oldStr;

    return *this;
}


//--------------------------------------------------------------------------------------------------
/// Returns a const ptr to the UTF-8 formatted string.
//--------------------------------------------------------------------------------------------------
const cdp_char_t* CDPString::utf8() const
{
    return m_str;
}

//--------------------------------------------------------------------------------------------------
/// Returns the number of bytes (length) of the string.
//--------------------------------------------------------------------------------------------------
cdp_size_t CDPString::byteCount() const
{
    if (m_str)
    {
        return strlen(m_str);
    }
    else
    {
        return 0;
    }
}

//--------------------------------------------------------------------------------------------------
/// Returns a string created from an UTF-8 array.
//--------------------------------------------------------------------------------------------------
CDPString CDPString::fromUtf8(const cdp_char_t* utf8Ptr)
{
    if (utf8Ptr)
    {
        size_t length = strlen(utf8Ptr);
        
        CDPString str;
        delete[] str.m_str;
        str.m_str = new cdp_char_t[length + 1];
        memcpy(str.m_str, utf8Ptr, length);
        str.m_str[length] = 0;
        return str;
    }
    else
    {
        return CDPString();
    }
}

//--------------------------------------------------------------------------------------------------
/// Returns a string created from an ASCII (char) string
/// 
/// Any ASCII value above 127 will be encoded to UTF-8
//--------------------------------------------------------------------------------------------------
CDPString CDPString::fromChar(const cdp_char_t* charPtr)
{
    if (!charPtr)
    {
        return CDPString();
    }

    // From http://www.codeguru.com/cpp/misc/misc/multi-lingualsupport/article.php/c10451
    static const unsigned int MASKBITS = 0x3F;
    static const unsigned int MASKBYTE = 0x80;
    static const unsigned int MASK2BYTES = 0xC0;

    size_t length = strlen(charPtr);
    cdp_char_t* tmpBuffer = new cdp_char_t[2*length + 1];
    size_t tmpSize = 0;

    for (size_t i = 0; i < length; i++)
    {
        const unsigned char uChar = static_cast<unsigned char>(charPtr[i]);
        unsigned int uc = static_cast<unsigned int>(uChar);

        // 0xxxxxxx
        if (uc < 0x80)
        {
            tmpBuffer[tmpSize++] = static_cast<char>(uc);
        }
        // 110xxxxx 10xxxxxx
        else if (uc < 0x800)
        {
            tmpBuffer[tmpSize++] = static_cast<char>(MASK2BYTES | (uc >> 6));
            tmpBuffer[tmpSize++] = static_cast<char>(MASKBYTE   | (uc & MASKBITS));
        }
    }

    tmpBuffer[tmpSize++] = 0;

    CDPString str = fromUtf8(tmpBuffer);
    delete[] tmpBuffer;

    return str;
}

