//##################################################################################################
//
//  Ceetron Data Provider Framework
//
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2018, Ceetron AS
//##################################################################################################

#include "TDLogging.h"
#include "TDProviderInterrogator.h"

#include "cdphlPluginManager.h"
#include "cdphlDataProviderFactory.h"
#include "cdphlDataProviderInfo.h"
#include "cdphlMetaData.h"
#include "cdphlGeometry.h"
#include "cdphlValidator.h"

#include "CDPPlugin.h"
#include "CDPDataProvider.h"
#include "CDPUtils.h"

#include <cstdlib>
#include <cstdio>


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
int main(int argc, char *argv[])
{
    TD_LOG_INFO("TestDriver starting...");
    TD_LOG_INFO("------------------------------");
    
    bool verboseMode = false;
    int reqArgCount = 3;
    
    if (argc > 1)
    {
        std::string arg1 = argv[1];
        if (arg1 == "-v")
        {
            verboseMode = true;
            reqArgCount = 4;
        }
    }
    
    if (argc < reqArgCount)
    {
        TD_LOG_ERROR("Not enough parameters specified.");
        TD_LOG_INFO("");
        TD_LOG_INFO("Usage: TestDriver [-v] <Plugin_file_base_name> <model_key>");
        TD_LOG_INFO("  -v : Verbose, list all data.");
        TD_LOG_INFO("");
        
        return EXIT_FAILURE;
    }

    // This should be the base file name of the plugin that is to be loaded
    // The base file name should NOT include the file extension (.dll, .so or .dylib)
    std::string pluginBaseFilePath = argv[reqArgCount - 2];
    std::string modelKey = argv[reqArgCount - 1];

    TD_LOG_INFO("Plugin base file name: " + pluginBaseFilePath);

    cdphl::PluginManager* pluginMgr = cdphl::PluginManager::instance();

    // Setup logger
    pluginMgr->setLogger(TDLogging::loggerInstance());

    if (!pluginMgr->loadPlugin(pluginBaseFilePath))
    {
        TD_LOG_ERROR("Failed to load and initialize the plugin: " + pluginBaseFilePath);
        return EXIT_FAILURE;
    }
    TD_LOG_INFO("Plugin successfully initialized");

    const size_t numProviders = pluginMgr->providerCount();
    TD_LOG_INFO(TDLogger::printf("Number of available providers: %d", static_cast<int>(numProviders)));
    for (size_t i = 0; i < numProviders; i++)
    {
        cdphl::DataProviderFactory* factory = pluginMgr->providerFactory(i);
        TD_LOG_INFO(TDLogger::printf("  provider(%d) idString: %s", static_cast<int>(i), factory->providerId().c_str()));
        
        cdphl::DataProviderInfo info;
        factory->getProviderInfo(&info);
        TD_LOG_INFO(TDLogger::printf("   - File descr (mask): %s (%s)",  info.fileDescription().utf8(), info.fileMask().utf8()));
    }


    for (size_t i = 0; i < pluginMgr->providerCount(); ++i)
    {
        cdphl::DataProviderFactory* factory = pluginMgr->providerFactory(i);
        const std::string providerIdString = factory->providerId();

        if (factory->isSupportedByProvider(CDPString::fromChar(modelKey.c_str())))
        {
            TD_LOG_INFO("");
            TD_LOG_INFO("Model key is supported by provider");
            TD_LOG_INFO(TDLogger::printf("Creating instance of provider: %s", providerIdString.c_str()));
            CDPDataProvider* provider = factory->createProviderInstance();
            if (!provider)
            {
                TD_LOG_ERROR("Could not create provider instance");
                return EXIT_FAILURE;
            }
            
//            TD_LOG_INFO("");
//            TD_LOG_INFO("");
//            TD_LOG_INFO(TDLogger::printf("Validating:  provider=%s  modelKey=%s", providerIdString.c_str(), modelKey.c_str()));
//            
//            cdphl::Validator validator(NULL);
//            validator.validate(provider, modelKey);

            TD_LOG_INFO("");
            TD_LOG_INFO("");
            TD_LOG_INFO(TDLogger::printf("Interrogating:  provider=%s  modelKey=%s", providerIdString.c_str(), modelKey.c_str()));
            TDProviderInterrogator interrogator(provider, verboseMode);
            interrogator.interrogate(modelKey);
        }
        else
        {
            TD_LOG_ERROR(TDLogger::printf("Provider '%s' does not support modelKey '%s'", providerIdString.c_str(), modelKey.c_str()));
        }
    }


    TD_LOG_INFO("");
    TD_LOG_INFO("TestDriver finished - SUCCESS!");

    return EXIT_SUCCESS;
}
