//##################################################################################################
//
//  Ceetron Data Provider Framework
//
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2018, Ceetron AS
//##################################################################################################

#include "TDProviderInterrogator.h"
#include "TDLogging.h"

#include "cdphlMetaData.h"
#include "cdphlGeometry.h"
#include "cdphlNodeGroup.h"
#include "cdphlElementGroup.h"
#include "cdphlSet.h"
#include "cdphlResult.h"
#include "cdphlResultValueGroup.h"
#include "cdphlTransformations.h"
#include "cdphlAssert.h"
#include "cdphlStringUtils.h"
#include "cdphlValidator.h"
#include "cdphlInitOptions.h"

#include "CDPDataProvider.h"
#include "CDPString.h"

#include <set>



//==================================================================================================
//
// 
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
TDProviderInterrogator::TDProviderInterrogator(CDPDataProvider* provider, bool verboseMode)
:   m_provider(provider),
    m_verboseMode(verboseMode)
{
    CDPHL_ASSERT(provider);
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
TDProviderInterrogator::~TDProviderInterrogator()
{

}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
bool TDProviderInterrogator::interrogate(std::string modelKey)
{
    cdphl::InitOptions options;
    const bool initOk = m_provider->init(CDPString::fromChar(modelKey.c_str()), options);
    if (!initOk)
    {
        TD_LOG_ERROR("Failed to initialize with modelKey: " + modelKey);
        m_provider->close();
        return false;
    }
    
    cdphl::Validator validator(TDLogging::loggerInstance());
    if (!validator.validate(m_provider, modelKey))
    {
        TD_LOG_ERROR("Validation of plugin FAILED!");
        return false;
    }
    
    TD_LOG_INFO("Validation passed.");
    TD_LOG_INFO("");

    TD_LOG_INFO("Querying meta data");
    cdphl::MetaData metaData;
    const bool metaOk = m_provider->getMetaData(&metaData);
    TD_LOG_INFO(std::string("Metadata retVal: ") + (metaOk ? "ok" : "fail"));

    TD_LOG_INFO("");
    TD_LOG_INFO("Listing geometryInfo");
    outputGeometryMetaData(metaData);

    TD_LOG_INFO("");
    TD_LOG_INFO("Listing stateInfo");
    outputStateMetaData(metaData);

    TD_LOG_INFO("");
    TD_LOG_INFO("Listing resultInfo");
    outputResultMetaData(metaData);

    TD_LOG_INFO("");
    TD_LOG_INFO("Listing setInfo");
    outputSetMetaData(metaData);

    TD_LOG_INFO("");
    TD_LOG_INFO("Interrogating all geometries");
    const bool showNodeCoordinates = m_verboseMode;
    const bool showElementConnectivities = m_verboseMode;
    interrogateAllGeometries(showNodeCoordinates, showElementConnectivities);

    TD_LOG_INFO("");
    TD_LOG_INFO("Interrogating all sets");
    const bool showSetItems = m_verboseMode;
    interrogateAllSets(showSetItems);

    TD_LOG_INFO("");
    TD_LOG_INFO("Interrogating all results");
    const bool showResultGroups = m_verboseMode;
    const bool showResultValuesInsideGroups = m_verboseMode;
    interrogateAllResults(showResultGroups, showResultValuesInsideGroups);

    m_provider->close();

    return true;
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void TDProviderInterrogator::interrogateAllGeometries(bool showNodeCoordinates, bool showElementConnectivities)
{
    cdphl::MetaData metaData;
    m_provider->getMetaData(&metaData);

    const size_t geoCount = metaData.geometryInfoCount();
    TD_LOG_INFO(TDLogger::printf("Geometry count: %d", geoCount));

    std::vector<cdp_int_t> allStateIds = getAllStateIds(metaData);
    if (allStateIds.size() == 0)
    {
        TD_LOG_ERROR("Cannot interrogate geometries, no states found!");
        return;
    }

    for (size_t ig = 0; ig < geoCount; ig++)
    {
        const cdp_size_t geometryIndex = ig;
        const cdphl::GeometryInfo& geoInfo = metaData.geometryInfo(ig);

        if (geoInfo.hasAdaptiveGeometry())
        {
            for (size_t is = 0; is < allStateIds.size(); is++)
            {
                const cdp_int_t stateId = allStateIds[is];
                cdphl::GeometrySpec spec(geometryIndex, stateId);
                cdphl::Geometry geo;
                const bool geoOk = m_provider->getGeometry(spec, &geo);
                if (!geoOk)
                {
                    TD_LOG_ERROR(TDLogger::printf("  Failed to get adaptive geometry with geometryIndex=%d stateId=%d", geometryIndex, stateId));
                    continue;
                }
                
                // Need to do this to complete the element definitions
                geo.computeDerivedData();

                TD_LOG_INFO(TDLogger::printf("  geometry (geometryIndex=%d  stateId=%d): eltGrpCount: %d, nodeGrpCount: %d", geometryIndex, stateId, geo.elementGroupCount(), geo.nodeGroupCount()));
                interrogateNodeGroupsInGeometry(geo, showNodeCoordinates);
                interrogateElementGroupsInGeometry(geo, showElementConnectivities);
            }
        }
        else
        {
            const cdp_int_t firstStateId = allStateIds[0];
            cdphl::GeometrySpec spec(geometryIndex, firstStateId);
            cdphl::Geometry geo;
            const bool geoOk = m_provider->getGeometry(spec, &geo);
            if (!geoOk)
            {
                TD_LOG_ERROR(TDLogger::printf("  Failed to get fixed geometry with geometryIndex=%d stateId=%d", geometryIndex, firstStateId));
                continue;
            }
            
            // Need to do this to complete the element definitions
            geo.computeDerivedData();

            TD_LOG_INFO(TDLogger::printf("  geometry (geometryIndex=%d  stateId=%d): eltGrpCount: %d, nodeGrpCount: %d", geometryIndex, firstStateId, geo.elementGroupCount(), geo.nodeGroupCount()));
            interrogateNodeGroupsInGeometry(geo, showNodeCoordinates);
            interrogateElementGroupsInGeometry(geo, showElementConnectivities);
        }
    }
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void TDProviderInterrogator::interrogateNodeGroupsInGeometry(const cdphl::Geometry& geo, bool showCoordinates)
{
    const size_t nodeGroupCount = geo.nodeGroupCount();
    for (size_t igrp = 0; igrp < nodeGroupCount; igrp++)
    {
        const cdphl::NodeGroup& nodeGrp = geo.nodeGroup(igrp);
        const size_t nodeCount = nodeGrp.nodeArr().size()/3;

        TD_LOG_INFO(TDLogger::printf("    nodeGrp(idx=%d)  nodeCount=%d", igrp, nodeCount));

        if (showCoordinates)
        {
            for (size_t in = 0; in < nodeCount; in++)
            {
                double x = nodeGrp.nodeArr()[3*in];
                double y = nodeGrp.nodeArr()[3*in + 1];
                double z = nodeGrp.nodeArr()[3*in + 2];
                TD_LOG_INFO(TDLogger::printf("      node(idx=%3d)  xyz=[%f %f %f]", in, x, y, z));
            }
        }
    }
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void TDProviderInterrogator::interrogateElementGroupsInGeometry(const cdphl::Geometry& geo, bool showElementConnectivities)
{
    const size_t eltGroupCount = geo.elementGroupCount();
    for (size_t igrp = 0; igrp < eltGroupCount; igrp++)
    {
        const cdphl::ElementGroup& eltGrp = geo.elementGroup(igrp);

        const size_t nodeGrpIdx = eltGrp.nodeGroupIndex();
        const size_t nodeCount = geo.nodeGroup(nodeGrpIdx).nodeArr().size()/3;

        std::vector<CDPElementType> elementTypeArr;
        std::vector<int> nodesPerElementArr;
        std::vector<int> elementNodeIndicesArr;

        const size_t subGroupDataCount = eltGrp.subGroupDataCount();
        for (size_t isub = 0; isub < subGroupDataCount; isub++)
        {
            const cdphl::ElementSubGroupData& subGroup = eltGrp.subGroupData(isub);
            size_t elementStartIdx = 0;
            for (size_t ie = 0; ie < subGroup.elementCount; ie++)
            {
                const CDPElementType elementType = subGroup.elementTypeArr.size() > 0 ? subGroup.elementTypeArr[ie] : subGroup.elementType;
                elementTypeArr.push_back(elementType);

                const int numElementNodes = subGroup.numElementNodes(ie);
                nodesPerElementArr.push_back(numElementNodes);

                for (size_t i = 0; i < static_cast<size_t>(numElementNodes); i++)
                {
                    elementNodeIndicesArr.push_back(subGroup.elementNodeArr[elementStartIdx + i]);
                }

                elementStartIdx += numElementNodes;
            }
        }

        const size_t elementCount = elementTypeArr.size();

        const std::set<int> referencedPartIdSet(eltGrp.elementPartIdArr().begin(), eltGrp.elementPartIdArr().end());
        const size_t numUniquePartIdsReferenced = referencedPartIdSet.size();

        TD_LOG_INFO(TDLogger::printf("    eltGrp(idx=%d)  elementCount=%d  nodeGrpIdx=%d  nodeCount=%d  numPartIdsReferenced=%d", igrp, elementCount, nodeGrpIdx, nodeCount, numUniquePartIdsReferenced));

        if (showElementConnectivities)
        {
            size_t elementStartIdx = 0;
            for (size_t ie = 0; ie < elementCount; ie++)
            {
                const CDPElementType elementType = elementTypeArr[ie];
                const size_t numElementNodes = static_cast<size_t>(nodesPerElementArr[ie]);

                const std::string elTypeStr = padLeft(elementTypeToString(elementType), 6);

                std::string partIdStr;
                if (eltGrp.elementPartIdArr().size() > 0)
                {
                    const cdp_int_t partId = eltGrp.elementPartIdArr()[ie];
                    partIdStr = TDLogger::printf(" partId=%d ", partId);
                }

                std::string elNodeIndicesStr;
                for (size_t i = 0; i < numElementNodes; i++)
                {
                    elNodeIndicesStr += cdphl::StringUtils::toStdString(elementNodeIndicesArr[elementStartIdx + i]) + std::string(" ");
                }

                elementStartIdx += numElementNodes;

                TD_LOG_INFO(TDLogger::printf("      el(idx=%3d)  type=%s(%d)  #nodes=%d %s nodeIndices=%s", ie, elTypeStr.c_str(), elementType, numElementNodes, partIdStr.c_str(), elNodeIndicesStr.c_str()));
            }
        }
    }
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void TDProviderInterrogator::interrogateAllResults(bool showResultGroups, bool showResultValuesInsideGroups)
{
    const bool showResultValues = showResultGroups && showResultValuesInsideGroups;

    cdphl::MetaData metaData;
    m_provider->getMetaData(&metaData);

    std::vector<cdp_int_t> allResultIds = getAllResultIds(metaData);
    std::vector<cdp_int_t> allStateIds = getAllStateIds(metaData);
    const size_t resultCount = allResultIds.size();
    const size_t stateCount = allStateIds.size();
    const size_t geoCount = metaData.geometryInfoCount();

    TD_LOG_INFO(TDLogger::printf("Result count: %d  (geoCount=%d stateCount=%d)", resultCount, geoCount, stateCount));
    for (size_t is = 0; is < stateCount; is++)
    {
        for (size_t ir = 0; ir < resultCount; ir++)
        {
            for (size_t ig = 0; ig < geoCount; ig++)
            {
                const cdp_int_t resultId = allResultIds[ir];
                const cdp_int_t stateId = allStateIds[is];
                const cdp_size_t geometryIndex = ig;

                cdphl::ResultSpec spec(geometryIndex, resultId, stateId);
                cdphl::Result res;
                const bool resOk = m_provider->getResult(spec, &res);
                if (!resOk)
                {
                    TD_LOG_ERROR(TDLogger::printf("  Failed to get result for stateId=%d resultId=%d geometryIndex=%d", stateId, resultId, geometryIndex));
                    continue;
                }

                const size_t groupCount = res.groupCount();
                TD_LOG_INFO(TDLogger::printf("  result (stateId=%d resultId=%d geometryIndex=%d): resultGrpCount: %d", stateId, resultId, geometryIndex, groupCount));

                if (showResultGroups)
                {
                    for (size_t igrp = 0; igrp < groupCount; igrp++)
                    {
                        const cdphl::ResultValueGroup& resGroup = res.group(igrp);
                        const size_t dimension = resGroup.valueDimension();
                        const size_t valueCount = resGroup.valueCount();
                        TD_LOG_INFO(TDLogger::printf("    resGrp(idx=%d)  dimension=%d  valueCount=%d", igrp, dimension, valueCount));

                        if (showResultValues)
                        {
                            for (size_t iv = 0; iv < valueCount; iv++)
                            {
                                std::string valueStr;
                                for (size_t i = 0; i < dimension; i++)
                                {
                                    valueStr += cdphl::StringUtils::toStdString(resGroup.resultValueArr()[iv*dimension + i]) + std::string(" ");
                                }

                                TD_LOG_INFO(TDLogger::printf("      idx=%3d:  %s", iv, valueStr.c_str()));
                            }
                        }
                    }
                }
            }
        }
    }
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void TDProviderInterrogator::interrogateAllSets(bool showSetItems)
{
    cdphl::MetaData metaData;
    m_provider->getMetaData(&metaData);

    const size_t setCount = metaData.setInfoCount();
    TD_LOG_INFO(TDLogger::printf("Set count: %d", setCount));
    for (size_t is = 0; is < setCount; is++)
    {
        const cdp_int_t setId = metaData.setInfo(is).id;

        cdphl::Set set;
        const bool setOk = m_provider->getSet(setId, &set);
        if (!setOk)
        {
            TD_LOG_ERROR(TDLogger::printf("  Failed to get set with setId=%d", setId));
            continue;
        }

        const size_t itemCount = set.itemCount();
        TD_LOG_INFO(TDLogger::printf("  set (setId=%d): itemCount: %d", setId, itemCount));

        if (showSetItems)
        {
            for (size_t i = 0; i < itemCount; i++)
            {
                const cdphl::SetItem& item = set.item(i);
                TD_LOG_INFO(TDLogger::printf("      idx=%3d:  stateId=%d  geoIdx=%d  eltGrpIdx=%2d  itemIdx=%d", i, item.stateId, item.geometryIndex, item.groupIndex, item.itemIndex));
            }
        }
    }
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void TDProviderInterrogator::outputGeometryMetaData(const cdphl::MetaData& metaData)
{
    const size_t geoCount = metaData.geometryInfoCount();
    TD_LOG_INFO(TDLogger::printf("Geometry count: %d", geoCount));
    for (size_t igeo = 0; igeo < geoCount; igeo++)
    {
        const cdphl::GeometryInfo& geoInfo = metaData.geometryInfo(igeo);
        const size_t eltGroupCount = geoInfo.elementGroupInfoCount();
        const bool hasAdaptiveGeometry = geoInfo.hasAdaptiveGeometry();

        TD_LOG_INFO(TDLogger::printf("  geometry(idx=%d): eltGroupCount=%d adaptiveGeometry=%s", igeo, eltGroupCount, (hasAdaptiveGeometry ? "yes" : "no")));

        for (size_t igrp = 0; igrp < eltGroupCount; igrp++)
        {
            const cdphl::ElementGroupInfo& eltGroupInfo = geoInfo.elementGroupInfo(igrp);
            TD_LOG_INFO(TDLogger::printf("    group(idx=%d): defPartName='%s' defPartId=%d", igrp, eltGroupInfo.defaultPartName().utf8(), eltGroupInfo.defaultPartId()));

            const size_t partCount = eltGroupInfo.partInfoCount();
            for (size_t ip = 0; ip < partCount; ip++)
            {
                const cdphl::PartInfo& partInfo = eltGroupInfo.partInfo(ip);
                TD_LOG_INFO(TDLogger::printf("      part(idx=%d): id=%d  name='%s'", ip, partInfo.id, partInfo.name.utf8()));
            }
        }
    }
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void TDProviderInterrogator::outputStateMetaData(const cdphl::MetaData& metaData)
{
    const size_t stateCount = metaData.stateInfoCount();
    TD_LOG_INFO(TDLogger::printf("State count: %d", stateCount));
    for (size_t is = 0; is < stateCount; is++)
    {
        const cdphl::StateInfo& stateInfo = metaData.stateInfo(is);
        TD_LOG_INFO(TDLogger::printf("  state(idx=%2d): id=%d  refValue=%f  name='%s' groupId=%d", is, stateInfo.id(), stateInfo.referenceValue(), stateInfo.name().utf8(), stateInfo.groupId()));
        
        if (stateInfo.groupId() >= 0)
        {
            bool found = false;
            const size_t stateGroupCount = metaData.stateGroupInfoCount();
            for (size_t j = 0; j < stateGroupCount; j++)
            {
                const cdphl::StateGroupInfo& stateGroupInfo = metaData.stateGroupInfo(j);
                if (stateGroupInfo.id == stateInfo.groupId())
                {
                    found = true;
                    break;
                }
            }
            
            if (!found)
            {
                TD_LOG_ERROR(TDLogger::printf("State id=%d (idx=%d) refers to stateGroup id=%d which does not exist!", is, stateInfo.id(), stateInfo.groupId()));
            }
        }
    }

    const size_t stateGroupCount = metaData.stateGroupInfoCount();
    TD_LOG_INFO(TDLogger::printf("State group: %d", stateGroupCount));
    for (size_t is = 0; is < stateGroupCount; is++)
    {
        const cdphl::StateGroupInfo& stateGroupInfo = metaData.stateGroupInfo(is);
        TD_LOG_INFO(TDLogger::printf("  stateGroup(idx=%2d): id=%d  name='%s'", is, stateGroupInfo.id, stateGroupInfo.name.utf8()));
    }
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void TDProviderInterrogator::outputResultMetaData(const cdphl::MetaData& metaData)
{
    const size_t resultCount = metaData.resultInfoCount();
    TD_LOG_INFO(TDLogger::printf("Result count: %d", resultCount));
    for (size_t ir = 0; ir < resultCount; ir++)
    {
        const cdphl::ResultInfo& resultInfo = metaData.resultInfo(ir);
        std::string typeStr = padLeft(resultTypeToString(resultInfo.type()), 12);
        std::string mappingStr = padLeft(resultMappingTypeToString(resultInfo.mapping()), 14);

        TD_LOG_INFO(TDLogger::printf("  result(idx=%2d): id=%d  type=%s(%d)  mapping=%s(%d)  name='%s'", ir, resultInfo.id(), typeStr.c_str(), resultInfo.type(), mappingStr.c_str(), resultInfo.mapping(), resultInfo.name().utf8()));
    }
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void TDProviderInterrogator::outputSetMetaData(const cdphl::MetaData& metaData)
{
    const size_t setCount = metaData.setInfoCount();
    TD_LOG_INFO(TDLogger::printf("Set count: %d", setCount));
    for (size_t is = 0; is < setCount; is++)
    {
        const cdphl::SetInfo& setInfo = metaData.setInfo(is);

        TD_LOG_INFO(TDLogger::printf("  set(idx=%2d): id=%d  type=%d  name='%s'", is, setInfo.id, setInfo.setType, setInfo.name.utf8()));
    }
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
std::vector<cdp_int_t> TDProviderInterrogator::getAllStateIds(const cdphl::MetaData& metaData)
{
    std::vector<cdp_int_t> allIds;

    const size_t stateCount = metaData.stateInfoCount();
    for (size_t i = 0; i < stateCount; i++)
    {
        const cdphl::StateInfo& stateInfo = metaData.stateInfo(i);
        allIds.push_back(stateInfo.id());
    }

    return allIds;
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
std::vector<cdp_int_t> TDProviderInterrogator::getAllResultIds(const cdphl::MetaData& metaData)
{
    std::vector<cdp_int_t> allIds;

    const size_t resultCount = metaData.resultInfoCount();
    for (size_t i = 0; i < resultCount; i++)
    {
        const cdphl::ResultInfo& resInfo = metaData.resultInfo(i);
        allIds.push_back(resInfo.id());
    }

    return allIds;
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
std::string TDProviderInterrogator::elementTypeToString(CDPElementType elementType)
{
    switch (elementType)
    {
        case CDP_ELT_UNKNOWN:              return "UNKNWN";
        case CDP_ELT_BEAMS:                return "BEAM";
        case CDP_ELT_BEAMS_3:              return "BEAM3";
        case CDP_ELT_TRIANGLES:            return "TRI";
        case CDP_ELT_TRIANGLES_6:          return "TRI6";
        case CDP_ELT_QUADS:                return "QUAD";
        case CDP_ELT_QUADS_8:              return "QUAD8";
        case CDP_ELT_QUADS_9:              return "QUAD9";
        case CDP_ELT_TETRAHEDRONS:         return "TET";
        case CDP_ELT_TETRAHEDRONS_10:      return "TET10";
        case CDP_ELT_HEXAHEDRONS:          return "HEX";
        case CDP_ELT_HEXAHEDRONS_20:       return "HEX20";
        case CDP_ELT_PENTAHEDRONS:         return "PENT";
        case CDP_ELT_PENTAHEDRONS_15:      return "PENT15";
        case CDP_ELT_POINTS:               return "POINT";
        case CDP_ELT_PYRAMIDS:             return "PYR";
        case CDP_ELT_PYRAMIDS_13:          return "PYR13";
        case CDP_ELT_POLYHEDRONS:          return "POLY";
    }

    return "UNDEF";
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
std::string TDProviderInterrogator::resultTypeToString(CDPResultType resultType)
{
    switch (resultType)
    {
        case CDP_RESTYPE_SCALAR:                    return "SCALAR";
        case CDP_RESTYPE_VECTOR:                    return "VECTOR";
        case CDP_RESTYPE_DISPLACEMENT:              return "DISPLACEMENT";
        case CDP_RESTYPE_STRESS_TENSOR:             return "STRESS_TENSOR";
        case CDP_RESTYPE_STRAIN_TENSOR:             return "STRAIN_TENSOR";
    }

    return "UNKNOWN";
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
std::string TDProviderInterrogator::resultMappingTypeToString(CDPResultMappingType mappingType)
{
    switch (mappingType)
    {
        case CDP_PER_NODE:              return "PER_NODE";
        case CDP_PER_ELEMENT:           return "PER_ELEMENT";
        case CDP_PER_ELEMENT_NODE:      return "PER_EL_NODE";
        case CDP_PER_ELEMENT_SURFACE:   return "PER_EL_SURFACE";
    }

    return "UNKNOWN";
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
std::string TDProviderInterrogator::padRight(const std::string& str, size_t width)
{
    if (str.size() < width)
    {
        return str + std::string(width - str.size(), ' ');
    }
    else
    {
        return str;
    }
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
std::string TDProviderInterrogator::padLeft(const std::string& str, size_t width)
{
    if (str.size() < width)
    {
        return std::string(width - str.size(), ' ') + str;
    }
    else
    {
        return str;
    }
}

